// Graphics library by ladyada/adafruit with init code from Rossum 
// MIT license

#include "N958G.h"


#define swap(a, b) { uint16_t t = a; a = b; b = t; }

#define	NOP					0
#define	SOFTWARE_RESET				1
#define	READ_DISPLAY_ID				4
#define	READ_DISPLAY_STATUS			9
#define	READ_DISPLAY_POWER_MODE			0xA
#define	READ_DISPLAY_MADCTL			0xb
#define	READ_DISPLAY_PIXEL_FORMAT	        0xc
#define	READ_DISPLAY_IMAGE_MODE			0xd
#define	READ_DISPLAY_SIGNAL_MODE		0xe
#define	READ_DISPLAY_SELF_DIAGNOSTICS 		0xf
#define	SLEEP_IN 				0x10
#define	SLEEP_OUT 				0x11
#define	PARTIAL_MODE_ON 			0x12
#define	NORMAL_DISPLAY_MODE_ON 			0x13
#define	DISPLAY_INVERSION_OFF 			0x20
#define	DISPLAY_INVERSION_ON 			0x21
#define	GAMMA_SET 				0x26
#define	DISPLAY_OFF 				0x28
#define	DISPLAY_ON 				0x29
#define	COLUMN_ADDRESS_SET 			0x2a
#define	ROW_ADDRESS_SET 			0x2b
#define	MEMORY_WRITE 				0x2c
#define	MEMORY_READ 				0x2e
#define	PARTIAL_AREA				0x30
#define	VERTICAL_SCROLLING_DEFINITION 		0x33
#define	TEARING_EFFECT_LINE_OFF 		0x34
#define	TEARING_EFFECT_LINE_ON 			0x35
#define	MEMORY_ACCESS_CONTROL 			0x36
#define	VERTICAL_SCROLLING_START_ADDRESS 	0x37
#define	IDLE_MODE_OFF 				0x38
#define	IDLE_MODE_ON 				0x39
#define	INTERFACE_PIXEL_FORMAT			0x3a
#define	WRITE_DISPLAY_BRIGHTNESS		0x51
#define	WRITE_CTRL_DISPLAY			0x53
#define	WRITE_CONTENT_ADAPTIVE_BRIGHTNESS 	0x55

// Colors

#define WHITE		0xffffff
#define BLACK		0x000000
#define GREEN		0x00ff00
#define YELLOW		0xffff00
#define CYAN		0x00ffff
#define RED		0xff0000
#define BLUE		0x0000ff
#define ORANGE 		0xFFA500

// Font Indices
#define FONT_LENGTH		0
#define FONT_FIXED_WIDTH	2
#define FONT_HEIGHT		3
#define FONT_FIRST_CHAR		4
#define FONT_CHAR_COUNT		5
#define FONT_WIDTH_TABLE	6


#define DELAY 0x80

#define isFixedWidthFont(font)  (FontRead(font+FONT_LENGTH) == 0 && FontRead(font+FONT_LENGTH+1) == 0))

N958G::N958G(const uint8_t cs, const uint8_t rs, const uint8_t wr, const uint8_t res, const uint8_t pwr, const uint32_t color) {
  _cs = cs;
  _rs = rs;
  _res = res;
  _wr = wr;
  _pwr = pwr;
  _backgroundColor=color;
}

void N958G::init(void)  {
  digitalWrite(_cs, LOW);	// chip select active
  digitalWrite(_pwr, HIGH);  // switch LCD power ON
  delay(100);

  digitalWrite(_res, HIGH);		// set Reset-Pin
  delay(5);
  digitalWrite(_res, LOW);		// clear Reset-Pin
  delay(50);
  digitalWrite(_res, HIGH);		// set Reset-Pin
  delay(50);
  
  writeCommand(SLEEP_OUT);
  delay(10);
  writeCommand(DISPLAY_INVERSION_OFF);
  writeCommand(IDLE_MODE_OFF);
  writeCommand(NORMAL_DISPLAY_MODE_ON);
  writeCommand(VERTICAL_SCROLLING_DEFINITION);
  writeData(0x00);
  writeData(0x00);
  writeData(0x01);
  writeData(0x40);
  writeData(0x00);
  writeData(0x00);
  writeCommand(VERTICAL_SCROLLING_START_ADDRESS);
  writeData(0x00);
  writeData(0x00);
  writeCommand(MEMORY_ACCESS_CONTROL);
  writeData(0x60);
  writeCommand(INTERFACE_PIXEL_FORMAT);
  writeData(0x77);
  writeCommand(WRITE_DISPLAY_BRIGHTNESS);
  writeData(0x00);
  writeCommand(WRITE_CTRL_DISPLAY);
  writeData(0x24);
  writeCommand(WRITE_CONTENT_ADAPTIVE_BRIGHTNESS);
  writeData(0x01);
  
  delay(125);
  writeCommand(DISPLAY_ON);
  
  setBacklight(70);
  digitalWrite(_cs, HIGH);
}

void N958G::setBacklight(const uint8_t percent)
{
  uint16_t value = percent;
  if(value > 100)  value = 100;
  writeCommand(WRITE_DISPLAY_BRIGHTNESS);
  writeData( (uint8_t)((value * 255) / 100) );
}


void N958G::setAddrWindow(const uint16_t x0, const uint16_t y0, const uint16_t x1, const uint16_t y1)
{
  
  writeCommand(COLUMN_ADDRESS_SET);	// x = 0 ... 239
  writeData(x0>>8);			// 0
  writeData(x0);			// xstart
  writeData(x1>>8);			// 0
  writeData(x1);			// xend
  writeCommand(ROW_ADDRESS_SET);        // y = 0 ... 319
  writeData(y0>>8);			// ystart >> 8
  writeData(y0);			// ystart & 0xff
  writeData(y1>>8);			// yend >>8
  writeData(y1);			// yend & 0xff
  writeCommand(MEMORY_WRITE);
}


void N958G::setBackgroundColor(const uint32_t color)
{
  _backgroundColor=color;
}

void N958G::drawPixel(const uint16_t x, const uint16_t y,const uint32_t color) {

  if ((x >= width) || (y >= height)) return;
  digitalWrite(_cs, LOW);

  setAddrWindow(x,y,x+1,y+1);

  writeData((color & 0x00FF0000) >> 16);	// R
  writeData((color & 0x0000FF00) >> 8);		// G
  writeData(color & 0x000000FF);		// B

  digitalWrite(_cs, HIGH);
}

void N958G::fillScreen(const uint32_t color)
{
  digitalWrite(_cs, LOW);

  setAddrWindow(0, 0, width-1, height-1);
  
  for (uint16_t x=0; x < width; x++) {
    for (uint16_t y=0; y < height; y++) {
      writeData((color & 0x00FF0000) >> 16);	// R
      writeData((color & 0x0000FF00) >> 8);	// G
      writeData(color & 0x000000FF);		// B  
    }
  }
  digitalWrite(_cs, HIGH);
}


// fill a circle
void N958G::fillCircle(const uint16_t x0, const uint16_t y0, const uint16_t r, const uint32_t color) {
  int16_t f = 1 - r;
  int16_t ddF_x = 1;
  int16_t ddF_y = -2 * r;
  int16_t x = 0;
  int16_t y = r;

  drawVerticalLine(x0, y0-r, 2*r+1, color);

  while (x<y) {
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f += ddF_y;
    }
    x++;
    ddF_x += 2;
    f += ddF_x;
  
    drawVerticalLine(x0+x, y0-y, 2*y+1, color);
    drawVerticalLine(x0-x, y0-y, 2*y+1, color);
    drawVerticalLine(x0+y, y0-x, 2*x+1, color);
    drawVerticalLine(x0-y, y0-x, 2*x+1, color);
  }
}

// draw a circle outline
void N958G::drawCircle(const uint16_t x0, const uint16_t y0, const uint16_t r, 
			const uint32_t color) {
  int16_t f = 1 - r;
  int16_t ddF_x = 1;
  int16_t ddF_y = -2 * r;
  int16_t x = 0;
  int16_t y = r;

  drawPixel(x0, y0+r, color);
  drawPixel(x0, y0-r, color);
  drawPixel(x0+r, y0, color);
  drawPixel(x0-r, y0, color);

  while (x<y) {
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f += ddF_y;
    }
    x++;
    ddF_x += 2;
    f += ddF_x;
  
    drawPixel(x0 + x, y0 + y, color);
    drawPixel(x0 - x, y0 + y, color);
    drawPixel(x0 + x, y0 - y, color);
    drawPixel(x0 - x, y0 - y, color);
    
    drawPixel(x0 + y, y0 + x, color);
    drawPixel(x0 - y, y0 + x, color);
    drawPixel(x0 + y, y0 - x, color);
    drawPixel(x0 - y, y0 - x, color);
  }
}

// draw a rectangle
void N958G::drawRect(const uint16_t x, const uint16_t y, const uint16_t w, const uint16_t h, 
		      const uint32_t color) {
  // smarter version
  drawHorizontalLine(x, y, w, color);
  drawHorizontalLine(x, y+h-1, w, color);
  drawVerticalLine(x, y, h, color);
  drawVerticalLine(x+w-1, y, h, color);
}

void N958G::fillRect(uint16_t x, uint16_t y, const uint16_t w, const uint16_t h, 
		      const uint32_t color) {
  // smarter version
  digitalWrite(_cs, LOW);

  setAddrWindow(x, y, x+w-1, y+h-1);

  // display controller writes first in X and increments Y after window width
  for (y=0; y < h; y++) {
    for (x=0; x < w; x++) {
      writeData((color & 0x00FF0000) >> 16);	// R
      writeData((color & 0x0000FF00) >> 8);	// G
      writeData(color & 0x000000FF);		// B   
    }
  }
  digitalWrite(_cs, HIGH);
}

void N958G::drawVerticalLine(const uint16_t x, const uint16_t y, uint16_t length, const uint32_t color)
{
  if (x >= width) return;
  if (y+length >= height) length = height-y-1;

  drawFastLine(x,y,length,color,1);
}

void N958G::drawHorizontalLine(const uint16_t x, const uint16_t y, uint16_t length, const uint32_t color)
{
  if (y >= height) return;
  if (x+length >= width) length = width-x-1;

  drawFastLine(x,y,length,color,0);
}

void N958G::drawFastLine(const uint16_t x, const uint16_t y, uint16_t length, 
			  const uint32_t color, const uint8_t rotflag)
{
  digitalWrite(_cs, LOW);

  if (rotflag) {
    setAddrWindow(x, y, x, y+length);
  } else {
    setAddrWindow(x, y, x+length, y+1);
  }
  // setup for data

  while (length--) {
      writeData((color & 0x00FF0000) >> 16);	// R
      writeData((color & 0x0000FF00) >> 8);	// G
      writeData(color & 0x000000FF);		// B   
  }
  digitalWrite(_cs, HIGH);
}


// bresenham's algorithm - thx wikpedia
void N958G::drawLine(int16_t x0, int16_t y0, int16_t x1, int16_t y1, 
		      const uint32_t color) {
  uint16_t steep = abs(y1 - y0) > abs(x1 - x0);
  if (steep) {
    swap(x0, y0);
    swap(x1, y1);
  }

  if (x0 > x1) {
    swap(x0, x1);
    swap(y0, y1);
  }

  uint16_t dx, dy;
  dx = x1 - x0;
  dy = abs(y1 - y0);

  int16_t err = dx / 2;
  int16_t ystep;

  if (y0 < y1) {
    ystep = 1;
  } else {
    ystep = -1;}

  for (; x0<=x1; x0++) {
    if (steep) {
      drawPixel(y0, x0, color);
    } else {
      drawPixel(x0, y0, color);
    }
    err -= dy;
    if (err < 0) {
      y0 += ystep;
      err += dx;
    }
  }
}


/**
 * Draw a rectangle with rounder corners
 *
 * @param x the x coordinate of the upper left corner of the rectangle
 * @param y the y coordinate of the upper left corner of the rectangle
 * @param width width of the rectangle
 * @param height height of the rectangle
 * @param radius ????
 * @param 16-bit color default BLACK 
 *
 * Draws a rectangle the same as DrawRect() but with rounded corners.
 * Radius is a value from 1 to half the height or width of the rectangle.
 * (what does that really mean?????)
 * Which is it, or is it a the smaller of the two?
 * FIXME FIXME need more description here.
 *
 * @see DrawRect()
 */

void N958G::drawRoundRect(const uint16_t x, const uint16_t y, const uint16_t width, const uint16_t height, 
                           const uint16_t radius, const uint32_t color=BLACK) 
{
  int16_t tSwitch; 
  uint16_t x1 = 0, y1 = radius;
  tSwitch = 3 - 2 * radius;
	
  while (x1 <= y1) {
    this->drawPixel(x+radius - x1, y+radius - y1, color);
    this->drawPixel(x+radius - y1, y+radius - x1, color);

    this->drawPixel(x+width-radius + x1, y+radius - y1, color);
    this->drawPixel(x+width-radius + y1, y+radius - x1, color);
	    
    this->drawPixel(x+width-radius + x1, y+height-radius + y1, color);
    this->drawPixel(x+width-radius + y1, y+height-radius + x1, color);

    this->drawPixel(x+radius - x1, y+height-radius + y1, color);
    this->drawPixel(x+radius - y1, y+height-radius + x1, color);

    if (tSwitch < 0) {
    	tSwitch += (4 * x1 + 6);
    } else {
      tSwitch += (4 * (x1 - y1) + 10);
      y1--;
    }
    x1++;
  }
	  	
  this->drawHorizontalLine(x+radius, y, width-(2*radius), color);		// top
  this->drawHorizontalLine(x+radius, y+height, width-(2*radius), color);	// bottom
  this->drawVerticalLine(x, y+radius, height-(2*radius), color);		// left
  this->drawVerticalLine(x+width, y+radius, height-(2*radius), color);	// right
}

void N958G::selectFont(const Font_t font)
{
	this->Font = font;
}

uint8_t N958G::getFontHeight(void)
{
  return( FontRead(this->Font+FONT_HEIGHT) );
}

uint16_t N958G::putChar(uint16_t x, const uint16_t y, uint8_t c, const uint32_t color=BLACK)
// returns x position after printed character
// function is a little bit speed optimized, with AT328p 16MHz: 
// 0.61ms for 5x7 character
// 1.26ms for a 7x15 character 
// 3.55ms for a 10x24 character
{
  if (this->Font == 0)
    return x; // no font selected

  /*
  * check for special character processing
  */
  if (c < 0x20) {  // special char, do nothing
    return x;
  }
	   
  uint8_t fontWidth = 0;
  uint8_t FontHeight = FontRead(this->Font+FONT_HEIGHT);
  uint8_t bytes = (FontHeight+7)/8; /* calculates FontHeight in rounded up bytes */

  uint8_t firstChar = FontRead(this->Font+FONT_FIRST_CHAR);
  uint8_t charCount = FontRead(this->Font+FONT_CHAR_COUNT);

  uint16_t index = 0;
  uint8_t thielefont;

  if (c < firstChar || c >= (firstChar+charCount)) {
    return x; // invalid char
  }
  c-= firstChar;

  if ( isFixedWidthFont(this->Font) {
    thielefont = 0;
    fontWidth = FontRead(this->Font+FONT_FIXED_WIDTH); 
    index = c*bytes*fontWidth+FONT_WIDTH_TABLE;
  }
  else  {
    // variable width font, read width data, to get the index
    thielefont = 1;
    /*
     * Because there is no table for the offset of where the data
     * for each character glyph starts, run the table and add up all the
     * widths of all the characters prior to the character we
     * need to locate.
     */
    for (uint8_t i=0; i<c; i++) {  
      index += FontRead(this->Font+FONT_WIDTH_TABLE+i);
    }
    /*
     * Calculate the offset of where the font data
     * for our character starts.
     * The index value from above has to be adjusted because
     * there is potentialy more than 1 byte per column in the glyph,
     * when the characgter is taller than 8 bits.
     * To account for this, index has to be multiplied
     * by the FontHeight in bytes because there is one byte of font
     * data for each vertical 8 pixels.
     * The index is then adjusted to skip over the font fontWidth data
     * and the font header information.
     */
     index = index*bytes+charCount+FONT_WIDTH_TABLE;
     /*
      * Finally, fetch the fontWidth of our character
      */
     fontWidth = FontRead(this->Font+FONT_WIDTH_TABLE+c);
  }  // end else

  // if character does not fit to the screen, write it at the end (no line wrap)
  if (x + fontWidth >= this->width)  {
    x=width-fontWidth;
  }

  digitalWrite(_cs, LOW);

  // last but not least, draw the character
  setAddrWindow(x, y, x+fontWidth-1, y+FontHeight-1);

  /*
   * Need to write to Display Memory in X direction
   * Y automatically increments due to set window size
   * Because GLCD fonts are organized in Y direction, need some ugly 
   * Font Data accesses to map the Font to the display memory
   * GLCD has stored the font in the following way:
   * Each byte is one X-step and eight Y-Steps. First X/Y pixel is given by the LSB
   * After all Bytes of one X line (font width) the next Y+8 pixel are coming
   * if less than 8-bit are used in a byte the MSB's are not used (at least for size 7)
   */

  for (uint8_t i=0; i<bytes; i++)	/* each vertical byte */
  { 
    uint16_t page = i*fontWidth; // page must be 16 bit to prevent overflow		
    uint8_t bitpnt=0x01;    // pointer to the bit to be drawn
    uint8_t nr_of_pixel=8;  // initial value for a byte
    const uint8_t* tmpFontPnt=this->Font+index+page;

    /* is it last byte of multibyte tall font? or a very small font? */
    uint8_t lastByte=( FontHeight > 8 && FontHeight < (i+1)*8 ) || (FontHeight < 8);
    if ( lastByte )
    /* is it last byte of multibyte tall font? or a very small font? */
    {
      nr_of_pixel=FontHeight-i*8;
      if (FontHeight<8)  // Workaround: bug in font-data for small fonts...
         bitpnt <<= (i+1)*8-FontHeight-1;        // remove unused bits
       else
         bitpnt <<= (i+1)*8-FontHeight;        
    }
    for (uint8_t p=0; p<nr_of_pixel; p++)  {   // p increments in y direction
      //uint8_t yp = i*8+p;
      for (uint8_t j=0; j<fontWidth; j++)  {   // write in x-direction
         uint8_t data = FontRead(tmpFontPnt+j);
         if (data & bitpnt)  {
           writeData((color & 0x00FF0000) >> 16);	// R
           writeData((color & 0x0000FF00) >> 8);	// G
           writeData(color & 0x000000FF);		// B   
           //drawPixel(x+j, y+yp, color);
         }
         else  {
           writeData((_backgroundColor & 0x00FF0000) >> 16);	// R
           writeData((_backgroundColor & 0x0000FF00) >> 8);	// G
           writeData(_backgroundColor & 0x000000FF);		// B  
           //drawPixel(x+j, y+yp, _backgroundColor);
         }
       }  // for j
       bitpnt <<= 1;
     }  // end  for p
  }  // end for i
  digitalWrite(_cs, HIGH);
  return (x+fontWidth); // valid char, return new x position
}

uint16_t N958G::putS(uint16_t x, const uint16_t y, const char *strin, const uint32_t color=BLACK)
{
  uint8_t first=1;
  uint8_t fontHeight = FontRead(this->Font+FONT_HEIGHT);
  char * str=(char *)strin;

  while(*str)  {
    if (!first)  {
      // space between characters
      drawVerticalLine(x,y,fontHeight, _backgroundColor);
      x++;
    }
    x=this->putChar(x, y, (uint8_t)*str, color);
    str++;
    first=0;
  }
  return(x);
}


/**
 * output a String class string
 *
 * @param str String class string
 *
 * Outputs all the characters in the string to the text area. 
 * See PutChar() for a full description of how characters are
 * written to the text area.
 *
 * @see PutChar()
 * @see Puts_P()
 * @see DrawString()
 * @see DrawString_P()
 * @see write()
 */

uint16_t  N958G::putS(uint16_t x, const uint16_t y, const String &str, const uint32_t color=BLACK)
{
  uint8_t fontHeight = FontRead(this->Font+FONT_HEIGHT);

  for (uint8_t i = 0; i < str.length(); i++)  {
    if (i!=0)  {
      // space between characters
      drawVerticalLine(x,y,fontHeight, _backgroundColor);
      x++;
    }
    x=this->putChar(x, y, str[i], color);
  }
  return(x);
}


/**
 * Returns the pixel width of a character
 *
 * @param c character to be sized
 *
 * @return The width in pixels of the given character
 * including any inter-character gap pixels following the character when the character is
 * rendered on the display.
 *
 * @note The font for the character is the most recently selected font.
 *
 * @see StringWidth()
 * @see StringWidth_P()
 */

uint8_t N958G::charWidth(uint8_t c)
{
  uint8_t width = 0;
	
  if(isFixedWidthFont(this->Font)  {
    width = FontRead(this->Font+FONT_FIXED_WIDTH)+1;  // there is 1 pixel pad here
  } 
  else  { 
    // variable width font 
    uint8_t firstChar = FontRead(this->Font+FONT_FIRST_CHAR);
    uint8_t charCount = FontRead(this->Font+FONT_CHAR_COUNT);
	
    // read width data
    if (c >= firstChar && c < (firstChar+charCount)) {
      c -= firstChar;
      width = FontRead(this->Font+FONT_WIDTH_TABLE+c)+1;
    }
  }	
  return width;
}

/**
 * Returns the pixel width of a string
 *
 * @param str pointer to string stored in RAM
 *
 * @return the width in pixels of the sum of all the characters in the
 * the string pointed to by str. 
 *
 * @see CharWidth()
 * @see StringWidth_P()
 */

uint16_t N958G::stringWidth(const char* str)
{
  uint16_t width = 0;
	
  while(*str != 0) {
    width += this->charWidth(*str++);
  }	
  return width;
}


/**
 * Returns the pixel width of a character
 *
		 * @param str String class string
 *
 * @return the width in pixels of the sum of all the characters in the
 * the string pointed to by str. 
 *
 * @see CharWidth()
 * @see StringWidth()
 */

uint16_t N958G::stringWidth(const String &str)
{
  uint16_t width = 0;

  for (uint16_t i = 0; i < str.length(); i++)
  {
    width += this->charWidth(str[i]);
  }
	
  return width;
}


// end of file
